#include <stdlib.h>
#include <time.h>
#include <stdbool.h>
#include <unistd.h>
#include <ctype.h>

#include "define.h"
#include "Hub.h"
#include "iniParser.h"
#include "utility.h"

#define CHECK_SUM_OFFSET 2
#define CODE_SIZE_OFFSET 0xFB
#define GET_CODE_SIZE(buf) (int)(*(buf + CODE_SIZE_OFFSET) * 1024)

static bool GetFlashIndex(int iHubIndex, int* flashidx);
static char PollingRDSRStatus(int iHubIndex);
static bool XromToBuf(int index, char* buf, int buf_size, bool bConfig);
static int GetToolString(int iHubIndex, usb_device_descriptor* dev_dsc);
static bool ReadChipVersionData(int iHubIndex, char* chip_data);
static int GetFwChipID(int i);
static int GetBinChipID(int iHubIndex, char* pBinBuf, int iBufSize);
static int GetFlashRWSize(int iFlashidx, int iHubIndex, int iChipID);
static bool ModifyHubIspMode(int iHubIndex, int iMode);  // 0: Enter 1:Exit
static bool WriteFlashStatus(int iHubIndex, int flashidx);
static bool UnprotectFlash(int iHubIndex, int flashidx, FLASH_OPERATE_CMD foc);
static bool EraseFlash(int iHubIndex, int flashidx, bool reOpen);
static bool CreateIdBuffer(int iChipID);

static bool GetFlashIndex(int iHubIndex, int* flashidx)
{
    int i, ret;
    char buffer[64]  = {0};
    bool status      = false;
    unsigned int tmp = 0, length = 0;
    unsigned int value = 0;

    for (i = 0; i < m_flashcount; i++)
    {
        memset(buffer, 0, sizeof(buffer));

        // check if dummy data exist.
        if (m_flashinfo[i][FLASH_INFO_RDID_DUMMY_ADDRESS] > 0)
        {
            value = 0x0001;
        }
        else
        {
            value = 0x0002;
        }

        tmp = m_flashinfo[i][FLASH_INFO_READ_FALSH_COMMAND];
        tmp <<= 8;
        value &= 0x00ff;
        value |= tmp;
        length = m_flashinfo[i][FLASH_INFO_READ_FLASH_COMMAND_LENGTH];

        ret = SendVendorCommand(m_fdev[iHubIndex],  // device handle
                                true,               // USB_DIR
                                GL_ISP_READ,        // request
                                value,              // value
                                0x0000,             // index
                                length,             // len
                                buffer);            // buf
        if (ret < 0)
        {
            printf("GetFlashIndex SendVendorCommand Error!\n");
            status = false;
            goto exit;
        }

        if (memcmp(buffer, m_flashinfo[i] + 2, m_flashinfo[i][FLASH_INFO_READ_FLASH_COMMAND_LENGTH]) == 0)
        {
            *flashidx = i;
            break;
        }
        else
        {
            if (i == m_flashcount - 1)
            {
                status = false;
                break;
            }
        }
    }

    status = true;
exit:
    return status;
}

static char PollingRDSRStatus(int iHubIndex)  // Check RDSR
{
    int ret          = 0;
    char rdsr_status = 0;

    ret = SendVendorCommand(m_fdev[iHubIndex],  // device handle
                            true,               // USB_DIR
                            GL_ISP_READ,        // request
                            0x0501,             // value
                            0x0000,             // index
                            1,                  // len
                            &rdsr_status);      // buf

    if (ret < 0)
    {
        printf("PollingRDSRStatus SendVendorCommand Error!\n");
        rdsr_status = -1;
    }
    else
    {
        rdsr_status &= BIT_OFFSET_0;
    }
    

    return rdsr_status;
}

static bool XromToBuf(int iHubIndex, char* buf, int buf_size, bool bConfig)
{
    int ret, i, j, loop, rem, flashidx = 0;
    int datasize = 64;
    bool status  = false;
    char data[64];
    unsigned int addr, index;

    printf("Entering ISP mode...\n");
    if (!ModifyHubIspMode(iHubIndex, 0))
    {
        status = false;
        goto exit;
    }

    status = GetFlashIndex(iHubIndex, &flashidx);
    if (status)
    {
        datasize = GetFlashRWSize(flashidx, iHubIndex, m_hubType[iHubIndex]);
    }
    else
    {
        printf("No Match Flash Type found!\n");
        status = false;
        goto exit;
    }

    usleep(100000);

    loop = buf_size / datasize;
    rem  = buf_size % datasize;
    if (rem)
        loop++;
    memset(buf, 0, buf_size);

    for (i = 0; i < loop; i++)
    {
        memset(data, 0, datasize);
        addr = i * datasize;

        if (bConfig)
        {
            index = addr + 0x0100;
        }
        else
        {
            index = addr;
        }

        for (j = 0; j < COMMAND_RETRY_TIMES; j++)
        {
            ret = SendVendorCommand(m_fdev[iHubIndex],  // device handle
                                    true,               // USB_DIR
                                    GL_ISP_READ,        // request
                                    0x0000,             // value
                                    index,              // index
                                    datasize,           // len
                                    data);              // buf

            if (ret >= 0)
            {
                memcpy(buf + addr, data, datasize);
                break;
            }
            else
            {
                if (j == COMMAND_RETRY_TIMES - 1)
                {
                    printf("read data Error!\n");
                    status = false;
                    goto exit;
                }
                else
                {
                    usleep(USLEEP_TIME);
                }
            }
        }
    }

    // exit isp mode
    if (!ModifyHubIspMode(iHubIndex, 1))
    {
        printf("exit isp mode error!\n");
        goto exit;
    }

    status = true;
exit:
    return status;
}

bool OpenDevices(void)
{
    bool status = false;
    int i       = 0;

    CloseDevices();

    SortUsbTree();

    if (m_devNum == 0)
    {
        printf("Cannot find GL device!\n");
        status = false;
        goto exit;
    }

    for (i = 0; i < m_devNum; i++)
    {
        m_fdev[i] = open(m_szDev[i], O_RDWR);

        if (m_fdev[i] == -1)
        {
            printf("Failed to open %s\n", m_szDev[i]);
            status = false;
            goto exit;
        }
        else
        {
            // printf("Device [%03d] %s\n", i+1, m_szDev[i]);
        }
    }

    status = true;
exit:
    return status;
}

void CloseDevices(void)
{
    int i = 0;

    for (i = 0; i < m_devNum; i++)
    {
        if (m_fdev[i] > 0)
        {
            close(m_fdev[i]);
        }
    }
}

static int GetToolString(int iHubIndex, usb_device_descriptor* dev_dsc)
{
    int ret    = 0;
    int status = false;
    int idx_S = 0, idx_D = 0;
    int StaticStringLen = 0x40, DynamicStringLen = 0x16;
    char szBuffer[0x40];
    unsigned char szDesc[0x20];
    usb_string_descriptor* pStrDesc = (usb_string_descriptor*)szBuffer;

    // printf("dev_dsc.bcdDevice:%x\n", dev_dsc->bcdDevice);
    if (dev_dsc->bcdUSB >= 0x300)
    {
        idx_S = 0x84;
        idx_D = 0x85;
    }
    else
    {
        idx_S = 0x81;
        idx_D = 0x82;
    }

    ret = GetDescriptor_String(m_fdev[iHubIndex], idx_S, StaticStringLen, pStrDesc);
    if (ret <= 0)
    {
        printf("GetDescriptor_String fail, status:%d\n", status);
        status = false;
        goto exit;
    }

    WstringToString(szDesc, 0x20, pStrDesc->wData, (pStrDesc->bLength - 2) / 2);

    m_GlStrDes[iHubIndex].StaticStr.StringVersion = szDesc[0];
    memcpy(m_GlStrDes[iHubIndex].StaticStr.MaskID, szDesc + 1, 7);
    memcpy(m_GlStrDes[iHubIndex].StaticStr.MaskHubIC, szDesc + 8, 6);
    memcpy(m_GlStrDes[iHubIndex].StaticStr.RunningHubID, szDesc + 14, 7);
    memcpy(m_GlStrDes[iHubIndex].StaticStr.RunningHubIC, szDesc + 21, 6);
    memcpy(m_GlStrDes[iHubIndex].StaticStr.FwVersion, szDesc + 27, 4);

    ret = GetDescriptor_String(m_fdev[iHubIndex], idx_D, DynamicStringLen, pStrDesc);
    if (ret <= 0)
    {
        printf("GetDescriptor_String fail, status:%d\n", status);
        status = false;
        goto exit;
    }

    WstringToString(szDesc, 0x20, pStrDesc->wData, (pStrDesc->bLength - 2) / 2);
    memcpy(&m_GlStrDes[iHubIndex].DynamicStr, szDesc, 10);

    status = true;
exit:
    return status;
}

static bool ReadChipVersionData(int iHubIndex, char* chip_data)
{
    int ret;
    bool status = false;
    char ch_temp;
    int hex_val   = -1;
    char data[64] = {0};
    int datasize  = 64;
    __u16 wValue  = 0;
    __u16 wIndex  = 0x01D0;
    int i         = 0;

    if (m_devNum <= 0 || iHubIndex < 0)
        goto exit;

    ret = SendVendorCommand(m_fdev[iHubIndex],  // device handle
                            true,               // USB_DIR
                            GL_ISP_READ,        // request
                            0x0000,             // value
                            0x0000,             // index
                            1,                  // len
                            data);              // buf
    if (ret < 0)
    {
        printf("Read Data Error!\n");
        status = false;
        goto exit;
    }

    ch_temp = m_GlStrDes[iHubIndex].DynamicStr.Bonding1;
    hex_val = ASCIIConvert(ch_temp);

    if (memcmp(m_GlStrDes[iHubIndex].StaticStr.MaskID, GL_7423_A01, 4) == 0 ||
        memcmp(m_GlStrDes[iHubIndex].StaticStr.MaskID, GL_8423_A01, 4) == 0)
    {
        if (hex_val & BIT_OFFSET_7)
            wValue = 0x1000;
    }
    else
    {
        if (hex_val & BIT_OFFSET_4)
        {
            wValue = 0x1000;
            wIndex |= 0x8000;
        }
    }

    ret = SendVendorCommand(m_fdev[iHubIndex],  // device handle
                            true,               // USB_DIR
                            GL_ISP_READ,        // request
                            wValue,             // value
                            wIndex,             // index
                            datasize,           // len
                            chip_data);         // buf
    if (ret < 0)
    {
        printf("Read Data Error!\n");
        status = false;
        goto exit;
    }

    for (i = 0; i < 64; i++)
        if (chip_data[i] == 0)
            chip_data[i] = 0xff;

    status = true;
exit:
    return status;
}

static int GetFwChipID(int iHubIndex)
{
    int hex_val   = -1;
    bool b7394_A  = false;
    int chipId    = CHIP_ID_UNKNOWN;
    int ret       = 0;
    char data[64] = {0};
    struct usb_device_descriptor dev_dsc;

    //
    ret = GetDescriptor_Device(m_fdev[iHubIndex], sizeof(dev_dsc), &dev_dsc);
    if (ret <= 0)
    {
        printf("GetDescriptor_Device fail, ret:%d\n", ret);
        chipId = CHIP_ID_UNKNOWN;
        goto exit;
    }

    m_devinfo[iHubIndex].vid       = dev_dsc.idVendor;
    m_devinfo[iHubIndex].pid       = dev_dsc.idProduct;
    m_devinfo[iHubIndex].bcdDevice = dev_dsc.bcdDevice;
    m_devinfo[iHubIndex].bcdUSB    = dev_dsc.bcdUSB;

    if (dev_dsc.bcdDevice == GL7344_MASK_CODE_FW_VERSION || dev_dsc.bcdDevice == GL7364_MASK_CODE_FW_VERSION ||
        dev_dsc.bcdDevice == GL7364_MASK_CODE_FW_VERSION_401 || dev_dsc.bcdDevice == GL7364_MASK_CODE_FW_VERSION_402 ||
        dev_dsc.bcdDevice == GL7394_MASK_CODE_FW_VERSION_500 || dev_dsc.bcdDevice == GL7394_MASK_CODE_FW_VERSION_501 ||
        dev_dsc.bcdDevice == GL7394_MASK_CODE_FW_VERSION_502 || dev_dsc.bcdDevice == GL7441_MASK_CODE_FW_VERSION_510 ||
        dev_dsc.bcdDevice == GL7441_MASK_CODE_FW_VERSION_511 || dev_dsc.bcdDevice == GL8423_MASK_CODE_FW_VERSION_600 ||
        dev_dsc.bcdDevice == GL7423_MASK_CODE_FW_VERSION_601 || dev_dsc.bcdDevice == GL7423_MASK_CODE_FW_VERSION_602 ||
        dev_dsc.bcdDevice == GL7423_MASK_CODE_FW_VERSION_610 ||
        dev_dsc.bcdDevice == GL7423A01_MASK_CODE_FW_VERSION_611 ||
        dev_dsc.bcdDevice == GL7423B02_MASK_CODE_FW_VERSION_612 ||
        dev_dsc.bcdDevice == GL7423C03_MASK_CODE_FW_VERSION_613 ||
        dev_dsc.bcdDevice == GL7423D04_MASK_CODE_FW_VERSION_614 ||
        dev_dsc.bcdDevice == GL7423E05_MASK_CODE_FW_VERSION_615 ||
        dev_dsc.bcdDevice == GL7423F06_MASK_CODE_FW_VERSION_616 ||
        dev_dsc.bcdDevice == GL7423G07_MASK_CODE_FW_VERSION_617 ||
        dev_dsc.bcdDevice == GL7423H08_MASK_CODE_FW_VERSION_618 ||
        dev_dsc.bcdDevice == GL7423I09_MASK_CODE_FW_VERSION_619 ||
        dev_dsc.bcdDevice == GL7448A001_MASK_CODE_FW_VERSION_628 ||
        dev_dsc.bcdDevice == GL7436_MASK_CODE_FW_VERSION_650 || dev_dsc.bcdDevice == GL7436_MASK_CODE_FW_VERSION_651 ||
        dev_dsc.bcdDevice == GL7436_MASK_CODE_FW_VERSION_652 || dev_dsc.bcdDevice == GL7436_MASK_CODE_FW_VERSION_653 ||
        dev_dsc.bcdDevice == GL7364_MASK_CODE_FW_VERSION_164)
    {
        m_bMaskRomCode[iHubIndex] = true;
    }

    ret = GetToolString(iHubIndex, &dev_dsc);
    if (ret <= 0)
    {
        printf("GetToolString fail, ret:%d\n", ret);
        chipId = CHIP_ID_UNKNOWN;
        goto exit;
    }

    if (m_bMaskRomCode[iHubIndex] == true)
    {
        GL_USB_DEVINFO devInfo;

        devInfo.bcdDev = dev_dsc.bcdDevice;
        chipId         = ReturnMaskCodeChipID(devInfo, m_GlStrDes[iHubIndex], iHubIndex);
    }
    else
    {
        ret = ReadChipVersionData(iHubIndex, data);
        if (ret == false)
        {
            printf("ReadChipVersionData fail, ret:%d\n", ret);
            chipId = CHIP_ID_UNKNOWN;
            goto exit;
        }

        if (strstr(data, GL7364_STRING) != NULL || strstr(data, GL7411_STRING) != NULL)
        {
            chipId = CHIP_ID_GL352X_21_SPI;
            goto exit;
        }
        else if (strstr(data, GL7394_STRING) != NULL || strstr(data, GL7441_STRING) != NULL)
        {
            hex_val = ASCIIConvert(m_GlStrDes[iHubIndex].DynamicStr.Bonding1);
            if (hex_val == 0)
            {
                // 7394 A ver. (Tool string version 0)
                chipId  = CHIP_ID_GL3523_SPI;
                b7394_A = true;
            }

            if (b7394_A == false)
            {
                if (m_GlStrDes[iHubIndex].StaticStr.StringVersion <= 0x31)
                {
                    if (hex_val & 0x04)
                        chipId = CHIP_ID_GL3523_SPI;
                    else
                        chipId = CHIP_ID_GL3523_TYPE_C;
                }
                else if (m_GlStrDes[iHubIndex].StaticStr.StringVersion >= 0x32)
                {
                    if ((hex_val & 0x01) == 0)
                        chipId = CHIP_ID_GL3523_SUPPORT_QC;
                    else if (hex_val & 0x08)
                        chipId = CHIP_ID_GL3523_SPI;
                    else
                        chipId = CHIP_ID_GL3523_TYPE_C;
                }
            }
            goto exit;
        }
        else if (strstr(data, GL7423_STRING) != NULL || strstr(data, GL8423_STRING) != NULL)
        {
            chipId = CHIP_ID_GL3590_SPI;
        }
        else if (strstr(data, GL7436_STRING) != NULL)
        {
            chipId = CHIP_ID_GL3510_SPI;
        }
        else if (strstr(data, GL7448_STRING) != NULL)
        {
            chipId = CHIP_ID_GL7000_SPI;
        }
    }

exit:
    return chipId;
}

static int GetBinChipID(int iHubIndex, char* pBinBuf, int iBufSize)
{
    int iChipId   = CHIP_ID_UNKNOWN;
    char data[11] = {0};

    if (pBinBuf == NULL || iBufSize < GL352X_21_CHIP_ID_OFFSET + 10)
        goto exit;

    memcpy(data, pBinBuf + GL352X_21_CHIP_ID_OFFSET, 10);

    if (strstr(data, GL7364_STRING) != NULL || strstr(data, GL7411_STRING) != NULL)
    {
        iChipId = CHIP_ID_GL352X_21_SPI;
    }
    else if (strstr(data, GL7394_STRING) != NULL || strstr(data, GL7441_STRING) != NULL)
    {
        // Check Type C flag @ offset 0x11D
        // 0x1D - Special Function5  bit -- 7:TYPEC_EN; 6:SOURCE_30A; 5:SOURCE_15A;
        //        4:P4_ISTEXTPC; 3:P3_ISTEXTPC; 2:P2_ISTEXTPC; 1:P1_ISTEXTPC; 0:P0_ISTEXTPC;
        //        TYPEC_EN=1  TypeC; TYPEC_EN = 0 standard
        if ((*(pBinBuf + 0x11D) & BIT_OFFSET_7) > 0)
        {
            iChipId = CHIP_ID_GL3523_TYPE_C;
        }
        else
        {
            iChipId = CHIP_ID_GL3523_SPI;
        }
    }
    else if (strstr(data, GL7423_STRING) != NULL || strstr(data, GL8423_STRING) != NULL)
    {
        iChipId = CHIP_ID_GL3590_SPI;
    }
    else if (strstr(data, GL7448_STRING) != NULL)
    {
        iChipId = CHIP_ID_GL7000_SPI;
    }
    else if (strstr(data, GL7436_STRING) != NULL)
    {
        iChipId = CHIP_ID_GL3510_SPI;
    }

exit:
    return iChipId;
}

void PrintFwVersion(void)
{
    int ret;
    struct usb_device_descriptor dev_dsc;

    for (int i = 0; i < m_devNum; i++)
    {
        ret = GetDescriptor_Device(m_fdev[i], sizeof(dev_dsc), &dev_dsc);
        if (ret <= 0)
        {
            printf("GetDescriptor_Device fail, ret:%d\n", ret);
            continue;
        }

        printf("GL hub %d version:%x\n", i, dev_dsc.bcdDevice);
    }
}

bool hub_erase_firmware(int iHubIndex)
{
    bool status  = false;
    int flashidx = 0;

    if (!HubIspInit())
    {
        printf("HubIspInit fail\n");
        goto exit;
    }

    if (!ModifyHubIspMode(iHubIndex, 0))
    {
        printf("enter isp mode error!\n");
        goto exit;
    }

    if (!GetFlashIndex(iHubIndex, &flashidx))
    {
        printf("Get flash type error!\n");
        goto exit;
    }
    // un-protect flash
    if (!UnprotectFlash(iHubIndex, flashidx, FOC_ERASE))
    {
        printf("un-protect status error!\n");
        goto exit;
    }
    // write status register
    if (!WriteFlashStatus(iHubIndex, flashidx))
    {
        printf("Write status error!\n");
        goto exit;
    }
    // erase flash by id
    printf("EraseFlash...\n");

    if (!EraseFlash(iHubIndex, flashidx, false))
    {
        printf("erase flash Error!\n");
        goto exit;
    }

    printf("EraseFlash ok, need re-plug Hub.\n");
    status = true;

exit:
    CloseDevices();

    return status;
}

bool hub_read_fw_version(void)
{
    int status = false;

    GetConfigFilePath();
    ReadTargetFromIni();

    status = OpenDevices();
    if (status == false)
    {
        printf("OpenDevices fail\n");
        goto exit;
    }

    PrintFwVersion();

    status = true;
exit:
    CloseDevices();

    return status;
}

bool hub_list_tree(void)
{
    int ret = true;

    system("lsusb -t");
    return ret;
}

bool hub_software_reset(int iHubIndex)
{
    int status = false;

    GetConfigFilePath();
    ReadTargetFromIni();

    status = OpenDevices();
    if (status == false)
    {
        printf("OpenDevices fail\n");
        goto exit;
    }

    SoftwareReset(iHubIndex);

    status = true;
exit:
    CloseDevices();

    return status;
}

bool hub_switch_speed(int iHubIndex, int iUsbSpeed)
{
    int status = false;

    GetConfigFilePath();
    ReadTargetFromIni();

    status = OpenDevices();
    if (status == false)
    {
        printf("OpenDevices fail\n");
        goto exit;
    }

    SwitchUsbSpeed(iHubIndex, iUsbSpeed);

    status = true;
exit:
    CloseDevices();

    return status;
}

bool hub_list_info(void)
{
    int status = false;
    int i;
    int iChipId;

    GetConfigFilePath();
    ReadTargetFromIni();

    status = OpenDevices();
    if (status == false)
    {
        printf("OpenDevices fail\n");
        goto exit;
    }

    for (i = 0; i < m_devNum; i++)
    {
        // Re-enter ISP mode
        if (!ModifyHubIspMode(i, 0))
        {
            printf("Hub %d Enter ISP mode failed!\n", i);
            continue;
        }

        iChipId = GetFwChipID(i);
        printf("GL hub %d : Chip Id(%X), VID_%04X, PID_%04X, REV_%04X\n", i, iChipId, m_devinfo[i].vid,
               m_devinfo[i].pid, m_devinfo[i].bcdDevice);
    }

    status = true;
exit:
    CloseDevices();

    return status;
}

bool HubIspInit(void)
{
    bool status = false;
    int i       = 0;

    // initlization.
    for (i = 0; i < GLHUB_MAX_NUM; i++)
    {
        m_fdev[i]    = -1;
        m_hubType[i] = CHIP_ID_UNKNOWN;
    }

    memset(m_eepFileName, 0, sizeof(m_eepFileName));
    memset(m_binFileName, 0, sizeof(m_binFileName));
    m_flashcount = 0;
    m_devNum     = 0;
    m_pBinBuf    = NULL;
    m_pConfigBuf = NULL;
    m_ispOK      = false;
    m_doingISP   = false;

    GetConfigFilePath();
    ReadTargetFromIni();

    status = OpenDevices();
    if (status == false)
    {
        printf("OpenDevices fail\n");
        goto exit;
    }

    status = ReadFlashInfoFromIni();
    if (status == false)
    {
        printf("Read flashinfo Error!\n");
        goto exit;
    }

    status = true;
exit:
    return status;
}

bool DoIsp(int iHubIndex)
{
    bool status = false;
    int i, j, loop;
    unsigned int rom_size;
    ulong addr, checksum = 0;
    char* bintemp         = NULL;
    unsigned int datasize = 64;
    int flashidx          = 0;
    char tmp;
    char* data          = 0;
    unsigned int wValue = 0;
    __u16 eeplen        = 0;
    char cFwVerLow;
    char cFwVerHigh;
    int ret       = 0;
    int nCodeSize = GET_CODE_SIZE(m_pBinBuf);

    printf("***** Start ISP Hub %d *****\n", iHubIndex + 1);

    if (!ModifyHubIspMode(iHubIndex, 0))
    {
        printf("Enter ISP mode failed!\n");
        goto exit;
    }
    else
    {
        printf("Entering ISP mode ok.\n");
    }

    m_hubType[iHubIndex] = GetFwChipID(iHubIndex);
    if (m_hubType[iHubIndex] == CHIP_ID_UNKNOWN)
    {
        printf("Unknown chip type!\n");
        goto exit;
    }

    if (m_hubType[iHubIndex] != m_BinType[iHubIndex])
    {
        printf("Bin file not match chip type!\n");
        goto exit;
    }

    switch (m_hubType[iHubIndex])
    {
        case CHIP_ID_GL352X_21_SPI:
        case CHIP_ID_GL352X_21_EEPROM:
            bintemp = new char[GL352X_21_ROM_SIZE + 2];
            memset(bintemp, 0xff, GL352X_21_ROM_SIZE);
            eeplen = GL352X_21_CONFIG_SIZE;
            break;

        case CHIP_ID_GL3523_SPI:
        case CHIP_ID_GL3523_TYPE_C:
        case CHIP_ID_GL3523_SUPPORT_QC:
        case CHIP_ID_GL3510_SPI:
            bintemp = new char[GL3523_ROM_SIZE + 2];
            memset(bintemp, 0xff, GL3523_ROM_SIZE);
            eeplen = GL3523_CONFIG_SIZE;
            break;

        case CHIP_ID_GL352X_21_EEPROM_LITE:
            bintemp = new char[GL352X_21_EEPROM_SIZE_LITE + 2];
            memset(bintemp, 0xff, GL352X_21_EEPROM_SIZE_LITE);
            break;

        case CHIP_ID_GL3590_SPI:
        case CHIP_ID_GL7000_SPI:
            bintemp = new char[nCodeSize];
            memset(bintemp, 0xff, nCodeSize);
            eeplen = GL3590_CONFIG_SIZE;
            break;

        default:
            printf("Unknown chip type!\n");
            goto exit;
    }

    if (m_bLoadEEP)
    {
        if (m_pConfigBuf == NULL)
        {
            printf("Configuration buffer is null!");
            goto exit;
        }

        // save F/W version data for patch EEP field.
        if (m_BinType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM_LITE)
        {
            cFwVerLow  = m_pBinBuf[FW_VER_L_EEPROM];
            cFwVerHigh = m_pBinBuf[FW_VER_H_EEPROM];
        }
        else
        {
            cFwVerLow  = m_pBinBuf[FW_VER_L_ADDR];
            cFwVerHigh = m_pBinBuf[FW_VER_H_ADDR];
        }

        memcpy(m_pBinBuf + EEP_ADDR_OFFSET, m_pConfigBuf, eeplen);

        // restore F/W version data.
        m_pBinBuf[FW_VER_L_ADDR] = cFwVerLow;
        m_pBinBuf[FW_VER_H_ADDR] = cFwVerHigh;
    }

    memcpy(bintemp, m_pBinBuf, m_2BinSize[iHubIndex]);

    // Put Container ID to F/W data
    if (!CreateIdBuffer(m_hubType[iHubIndex]))
    {
        printf("CreateIdBuffer fail!\n");
        goto exit;
    }

    switch (m_hubType[iHubIndex])
    {
        case CHIP_ID_GL352X_21_SPI:
        case CHIP_ID_GL352X_21_EEPROM:
            memcpy(bintemp + GL3521_CONTAINER_ID_OFFSET, m_containerID, BYTE_CONTAINER_ID_LENGTH);
            break;

        case CHIP_ID_GL3523_SPI:
        case CHIP_ID_GL3523_TYPE_C:
        case CHIP_ID_GL3523_SUPPORT_QC:
        case CHIP_ID_GL3590_SPI:
        case CHIP_ID_GL7000_SPI:
            memcpy(bintemp + GL3521_CONTAINER_ID_OFFSET, m_containerID, BYTE_CONTAINER_ID_LENGTH);
            break;

        case CHIP_ID_GL352X_21_EEPROM_LITE:
            memcpy(bintemp + GL3521_EEP_CONTAINER_ID_OFFSET, m_containerID, BYTE_CONTAINER_ID_LENGTH);
            break;

        case CHIP_ID_GL3510_SPI:
            memcpy(bintemp + GL3510_CONTAINER_ID_OFFSET, m_containerID, BYTE_CONTAINER_ID_LENGTH);
            break;
    }

    // calculate checksum
    switch (m_hubType[iHubIndex])
    {
        case CHIP_ID_GL352X_21_SPI:
        case CHIP_ID_GL352X_21_EEPROM:
            checksum = calc_checksum(bintemp, GL352X_21_CHECKSUN_OFFSET);
            break;

        case CHIP_ID_GL3523_SPI:
        case CHIP_ID_GL3523_TYPE_C:
        case CHIP_ID_GL3523_SUPPORT_QC:
        case CHIP_ID_GL3510_SPI:
            checksum = calc_checksum(bintemp, GL3523_CHECKSUN_OFFSET);
            break;

        case CHIP_ID_GL352X_21_EEPROM_LITE:
            checksum = calc_checksum(bintemp, GL352X_21_EEP_CHECKSUN_OFFSET);
            break;

        case CHIP_ID_GL3590_SPI:
        case CHIP_ID_GL7000_SPI:
            checksum = calc_checksum(bintemp, nCodeSize - CHECK_SUM_OFFSET);
            break;
    }

    if (!GetFlashIndex(iHubIndex, &flashidx))
    {
        printf("No Match Flash found!\n");
        goto exit;
    }

    printf("Matched flash index %d.\n", flashidx);

    m_iEraseDelay[iHubIndex] = m_flashinfo[flashidx][FLASH_INFO_ERASE_DELAY_TIME] * 100 * 1000;
    m_iWriteDelay[iHubIndex] = m_flashinfo[flashidx][FLASH_INFO_WRITE_DELAY_TIME] * 1000;

    //////////////////////////////////////////////////////////////////////////
    // Check SPI Flash
    //////////////////////////////////////////////////////////////////////////
    if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM_LITE)
    {
        rom_size = GL352X_21_EEPROM_SIZE_LITE;
        datasize = GL352X_21_EEP_TRANSFER_LENGTH_LITE;
    }
    else if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM)
    {
        rom_size = GL352X_21_EEPROM_SIZE;
        datasize = GL352X_21_EEP_TRANSFER_LENGTH;
    }
    else
    {
        if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_SPI)
        {
            rom_size = GL352X_21_ROM_SIZE;
        }
        else if (m_hubType[iHubIndex] == CHIP_ID_GL3523_SPI || m_hubType[iHubIndex] == CHIP_ID_GL3523_TYPE_C ||
                 m_hubType[iHubIndex] == CHIP_ID_GL3523_SUPPORT_QC || m_hubType[iHubIndex] == CHIP_ID_GL3510_SPI)
        {
            rom_size = GL3523_ROM_SIZE;
        }
        else if (m_hubType[iHubIndex] == CHIP_ID_GL3590_SPI || m_hubType[iHubIndex] == CHIP_ID_GL7000_SPI)
        {
            rom_size = nCodeSize;
        }

        datasize = GetFlashRWSize(flashidx, iHubIndex, m_hubType[iHubIndex]);

        // un-protect flash
        if (!UnprotectFlash(iHubIndex, flashidx, FOC_ERASE))
        {
            printf("un-protect status error!\n");
            goto exit;
        }

        printf("un-protect status ok\n");

        // write status register
        if (!WriteFlashStatus(iHubIndex, flashidx))
        {
            printf("Write status error!\n");
            goto exit;
        }

        printf("Write status ok\n");

        printf("EraseFlash...\n");
        // erase flash by id
        if (!EraseFlash(iHubIndex, flashidx, false))
        {
            printf("erase flash Error!\n");
            goto exit;
        }

        printf("EraseFlash ok\n");
    }

    data = new char[datasize];

    // write checksum.
    tmp = (checksum >> 8) & 0x00ff;
    memcpy(bintemp + rom_size - 2, &tmp, 1);
    tmp = checksum & 0x00ff;
    memcpy(bintemp + rom_size - 1, &tmp, 1);

    // calculate loop times.
    // Reset loop counter for EEPROM
    if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM_LITE || m_hubType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM)
    {
        loop = rom_size / datasize;
    }
    else
    {
        loop = m_2BinSize[iHubIndex] / datasize;
        if (m_2BinSize[iHubIndex] % datasize)
            loop++;

        if (m_hubType[iHubIndex] != CHIP_ID_GL3590_SPI && m_hubType[iHubIndex] != CHIP_ID_GL7000_SPI)
        {
            // for checksum.
            if (rom_size - m_2BinSize[iHubIndex] >= datasize)
                loop++;
            // for extra data.
            loop++;
        }
    }

    printf("Updating firmware...\n");

    //////////////////////////////////////////////////////////////////////////
    // Write
    //////////////////////////////////////////////////////////////////////////
    if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM_LITE)
    {
        wValue = 0x0200;
    }
    else if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM)
    {
        wValue = 0x0100;
    }
    else if (m_flashinfo[flashidx][FLASH_INFO_AAI_MODE_LOW])
    {
        tmp    = m_flashinfo[flashidx][FLASH_INFO_AAI_MODE_LOW];
        wValue = tmp;
        tmp    = m_flashinfo[flashidx][FLASH_INFO_AAI_MODE_HIGH];
        wValue |= (tmp << 8);
    }

    for (i = 0; i < loop; i++)
    {
        if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM_LITE || m_hubType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM ||
            m_hubType[iHubIndex] == CHIP_ID_GL3590_SPI || m_hubType[iHubIndex] == CHIP_ID_GL7000_SPI)
        {
            memcpy(data, bintemp + i * datasize, datasize);
            addr = i * datasize;
        }
        else
        {
            if (i == loop - 1)
            {
                // extra data
                if (datasize == 1)  // one byte checksum...
                {
                    if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_SPI)
                    {
                        memcpy(data, bintemp + GL352X_21_ROM_SIZE - datasize, datasize);
                        addr = GL352X_21_ROM_SIZE - datasize;
                    }
                    else if (m_hubType[iHubIndex] == CHIP_ID_GL3523_SPI ||
                             m_hubType[iHubIndex] == CHIP_ID_GL3523_TYPE_C ||
                             m_hubType[iHubIndex] == CHIP_ID_GL3523_SUPPORT_QC)
                    {
                        memcpy(data, bintemp + GL3523_ROM_SIZE - datasize, datasize);
                        addr = GL3523_ROM_SIZE - datasize;
                    }
                    else if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM)
                    {
                        memcpy(data, bintemp + GL352X_21_EEPROM_SIZE - datasize, datasize);
                        addr = GL352X_21_EEPROM_SIZE - datasize;
                    }
                    else if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM_LITE)
                    {
                        memcpy(data, bintemp + GL352X_21_EEPROM_SIZE_LITE - datasize, datasize);
                        addr = GL352X_21_EEPROM_SIZE_LITE - datasize;
                    }
                }
            }
            else if (i == loop - 2)
            {
                // checksum
                if (datasize == 1)
                {
                    if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_SPI)
                    {
                        memcpy(data, bintemp + GL352X_21_ROM_SIZE - datasize - 1, datasize);
                        addr = GL352X_21_ROM_SIZE - datasize - 1;
                    }
                    else if (m_hubType[iHubIndex] == CHIP_ID_GL3523_SPI ||
                             m_hubType[iHubIndex] == CHIP_ID_GL3523_TYPE_C ||
                             m_hubType[iHubIndex] == CHIP_ID_GL3523_SUPPORT_QC)
                    {
                        memcpy(data, bintemp + GL3523_ROM_SIZE - datasize - 1, datasize);
                        addr = GL3523_ROM_SIZE - datasize - 1;
                    }
                    else if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM)
                    {
                        memcpy(data, bintemp + GL352X_21_EEPROM_SIZE - datasize - 1, datasize);
                        addr = GL352X_21_EEPROM_SIZE - datasize - 1;
                    }
                    else if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM_LITE)
                    {
                        memcpy(data, bintemp + GL352X_21_EEPROM_SIZE_LITE - datasize - 1, datasize);
                        addr = GL352X_21_EEPROM_SIZE_LITE - datasize - 1;
                    }
                }
                else
                {
                    if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_SPI)
                    {
                        memcpy(data, bintemp + GL352X_21_ROM_SIZE - datasize, datasize);
                        addr = GL352X_21_ROM_SIZE - datasize;
                    }
                    else if (m_hubType[iHubIndex] == CHIP_ID_GL3523_SPI ||
                             m_hubType[iHubIndex] == CHIP_ID_GL3523_TYPE_C ||
                             m_hubType[iHubIndex] == CHIP_ID_GL3523_SUPPORT_QC)
                    {
                        memcpy(data, bintemp + GL3523_ROM_SIZE - datasize, datasize);
                        addr = GL3523_ROM_SIZE - datasize;
                    }
                    else if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM)
                    {
                        memcpy(data, bintemp + GL352X_21_EEPROM_SIZE - datasize, datasize);
                        addr = GL352X_21_EEPROM_SIZE - datasize;
                    }
                    else if (m_hubType[iHubIndex] == CHIP_ID_GL352X_21_EEPROM_LITE)
                    {
                        memcpy(data, bintemp + GL352X_21_EEPROM_SIZE_LITE - datasize, datasize);
                        addr = GL352X_21_EEPROM_SIZE_LITE - datasize;
                    }
                }
            }
            else
            {
                // normal
                memcpy(data, bintemp + i * datasize, datasize);
                addr = i * datasize;
            }
        }

        // un-protect flash
        if (!UnprotectFlash(iHubIndex, flashidx, FOC_WRITE))
        {
            printf("un-protect write (%04x) status error!\n", (unsigned int)addr);
            goto exit;
        }

        for (j = 0; j < COMMAND_RETRY_TIMES; j++)
        {
            ret = SendVendorCommand(m_fdev[iHubIndex],  // device handle
                                    false,              // USB_DIR
                                    GL_ISP_WRITE,       // request
                                    wValue,             // value
                                    addr,               // index
                                    datasize,           // len
                                    data);              // buf

            if (ret >= 0)
            {
                printf("Write complete %u %%..\r", i * 100 / loop);
                break;
            }
            else
            {
                if (j == COMMAND_RETRY_TIMES - 1)
                {
                    printf("write data ERROR! address %04x\n", (unsigned int)addr);
                    goto exit;
                }
                else
                {
                    usleep(USLEEP_TIME);
                    continue;
                }
            }
        }
    }

    // Read & Compare
    printf("Verifying firmware...\n");

    loop = rom_size / datasize;
    memset(data, 0, datasize);
    wValue = 0x0000;

    switch (m_hubType[iHubIndex])
    {
        case CHIP_ID_GL352X_21_EEPROM:
            wValue = 0x0100;
            break;

        case CHIP_ID_GL352X_21_EEPROM_LITE:
            wValue = 0x0200;
            break;
    }

    for (i = 0; i < loop; i++)
    {
        addr = i * datasize;

        for (j = 0; j < COMMAND_RETRY_TIMES; j++)
        {
            ret = SendVendorCommand(m_fdev[iHubIndex],  // device handle
                                    true,               // USB_DIR
                                    GL_ISP_READ,        // request
                                    wValue,             // value
                                    addr,               // index
                                    datasize,           // len
                                    data);              // buf

            if (ret >= 0)
            {
                break;
            }
            else
            {
                if (j == COMMAND_RETRY_TIMES - 1)
                {
                    printf("read data ERROR! address %04X", (unsigned int)addr);
                    goto exit;
                }
                else
                {
                    usleep(USLEEP_TIME);
                    continue;
                }
            }
        }

        if (memcmp(data, bintemp + i * datasize, datasize))
        {
            printf("compare data ERROR! address %04X\n", (unsigned int)addr);
            goto exit;
        }
    }

    // Exit ISP Mode
    if (!ModifyHubIspMode(iHubIndex, 1))
    {
        printf("Exit ISP mode failed!\n");
        goto exit;
    }
    else
    {
        printf("Exit ISP mode ok.\n");
    }

    status = true;

exit:
    if (bintemp)
        delete (bintemp);

    if (data)
        delete (data);

    if (m_pBinBuf)
        free(m_pBinBuf);

    if (m_pEepBuf)
        free(m_pEepBuf);

    if (status)
    {
        printf("***** ISP successful, need re-plug Hub. *****\n");
    }
    else
    {
        printf("***** ISP failed! *****\n");
    }
    

    return status;
}

static int GetFlashRWSize(int iFlashidx, int iHubIndex, int iChipID)
{
    // for legacy table trf_size = 0x40/0x32
    // for V2 table flexible optimal value is applied
    // GL7344 Write size: limited max 64 bytes only
    // GL7364/7411/7394-A01 Write size: limited max 128 bytes only
    // 7394-B02 Write size: limited max 256 bytes only

    int trf_size = m_flashinfo[iFlashidx][FLASH_INFO_FLASH_WRITE_LENGTH];  // max write size for specific flash type
    int AAI_mode = m_flashinfo[iFlashidx][FLASH_INFO_AAI_MODE_LOW];

    if (m_flashinfo[iFlashidx][FLASH_INFO_FLASH_WRITE_LENGTH] < 0x40)
        return m_flashinfo[iFlashidx][FLASH_INFO_FLASH_WRITE_LENGTH];

    // patch GL7394-B02 Mask Code Bug, flash support AAI mode programming not work.
    // apply one byte programming to work around
    if (AAI_mode && (iChipID == CHIP_ID_GL3523_SPI || iChipID == CHIP_ID_GL3523_TYPE_C))
    {
        if (memcmp(m_GlStrDes[iHubIndex].StaticStr.MaskID, GL_7394_B02, 7) == 0 &&
            m_bMaskRomCode[iHubIndex])  // running mask code
            return 1;
    }

    if (trf_size > 0x40)
    {
        if (m_devinfo[iHubIndex].bcdUSB >= 0x300)
        {
            if (trf_size > 0x80)
            {
                trf_size -= 0x80;
                trf_size *= 128;

                if (trf_size > 128 && trf_size < 512)
                {  // legacy chip support max 128 bytes write size only
                    switch (iChipID)
                    {
                        case CHIP_ID_UNKNOWN:
                            trf_size = 64;  // GL7344 transfer max 64 bytes only
                            break;          // and should not go this way beacuse bug
                            // can through USB 2.0 command pipe only

                        case CHIP_ID_GL352X_21_SPI:
                            trf_size = 128;
                            break;

                        case CHIP_ID_GL3590_SPI:
                            trf_size = 256;
                            break;

                        case CHIP_ID_GL3523_SPI:
                        case CHIP_ID_GL3523_TYPE_C:
                        case CHIP_ID_GL3510_SPI:
                        {
                            if (memcmp(m_GlStrDes[iHubIndex].StaticStr.MaskID, GL_7394_A01, 7))  // Non GL7394-A01 Chip
                            {
                                trf_size = 256;
                            }
                            else
                            {
                                trf_size = 128;
                            }

                            break;
                        }
                    }
                }

                if (trf_size >= 512)
                    trf_size = 512;  // USB3.0 command pipe teransfer max. 512 bytes

                return trf_size;
            }
            else
                return trf_size;
        }
        else              // USB2.0 command pipe teransfer max. 64 bytes
            return 0x40;  // High Speed / GL7344 Go This Way (because hw bug, force to go through USB2.0 command pipe)
    }
    else  // flash write size specify <= 64 bytes
        return trf_size;
}

static bool WriteFlashStatus(int iHubIndex, int flashidx)
{
    bool status = false;
    char tmp    = 0;
    int i, ret;

    tmp = m_flashinfo[flashidx][FLASH_INFO_EWSR];
    if (tmp == 0)
    {
        // EWSR = 0, skip WriteFlashStatus.
        // printf("EWSR = 0\n");
        status = true;
        goto exit;
    }

    for (i = 0; i < COMMAND_RETRY_TIMES; i++)
    {
        ret = SendVendorCommand(m_fdev[iHubIndex],  // device handle
                                false,              // USB_DIR
                                GL_ISP_SW,          // request
                                (tmp << 8) + 1,     // value
                                0x0000,             // index
                                0x0000,             // len
                                NULL);              // buf

        if (ret >= 0)
        {
            break;
        }
        else
        {
            usleep(USLEEP_TIME);
        }
    }

    if (ret < 0)
    {
        printf("Write Flash Status Error!\n");
        goto exit;
    }

    status = true;
exit:
    return status;
}

static bool ModifyHubIspMode(int iHubIndex, int iMode)  // 0: Enter 1:Exit
{
    bool status = false;
    int i;
    int ret      = 0;
    __u16 wValue = 0;
    unsigned int startctTime, tsTime;

    if (iMode == 0)
        wValue = 1;

    for (i = 0; i < COMMAND_RETRY_TIMES; i++)
    {
        ret = SendVendorCommand(m_fdev[iHubIndex],  // device handle
                                false,              // USB_DIR
                                GL_ISP_SW,          // request
                                wValue,             // value
                                0x0000,             // index
                                0x0000,             // len
                                NULL);              // buf

        if (ret >= 0)
        {
            break;
        }
        else
        {
            usleep(USLEEP_TIME);
        }
    }

    if (ret < 0)
    {
        printf("Modify ISP Mode(%d) Error!\n", iMode);
        goto exit;
    }

    startctTime = clock();
    tsTime      = 0;

    while (PollingRDSRStatus(iHubIndex))
    {
        if (tsTime > m_iEraseDelay[iHubIndex] && tsTime > 1500)
            goto exit;

        usleep(50000);
        tsTime = (clock() - startctTime);
    }

    status = true;
exit:
    return status;
}

static bool UnprotectFlash(int iHubIndex, int flashidx, FLASH_OPERATE_CMD foc)
{
    bool status     = false;
    int cmd_num     = 0;
    int ret         = 0;
    __u16 wValue;
    int i, j;

    if (foc == FOC_ERASE)
    {
    }
    else if (foc == FOC_WRITE)
    {
        if (m_GlStrDes[iHubIndex].StaticStr.StringVersion >= TOOL_STRING_VERSION_3 ||
            (m_flashinfo[flashidx][FLASH_INFO_UN_PROTECT_FLAG] & 0x60) <= 0)
            return true;

        if (m_flashinfo[flashidx][FLASH_INFO_UN_PROTECT_FLAG] & BIT_OFFSET_7)
        {
            // Enter ISP mode first.
            if (!ModifyHubIspMode(iHubIndex, 0))
            {
                printf("Enter ISP mode failed!\n");
                goto exit;
            }
        }
    }

    cmd_num = (m_flashinfo[flashidx][FLASH_INFO_UN_PROTECT_FLAG] & 0x0f);
    if (cmd_num > 4)
        cmd_num = 4;  // maximum = 4 commands

    for (j = 0; j < cmd_num; j++)
    {
        if (FLASH_INFO_UN_PROTECT_COMMAND_1 + j > FALSH_INFO_COUNT)
            break;

        wValue = m_flashinfo[flashidx][FLASH_INFO_UN_PROTECT_COMMAND_1 + j];
        wValue <<= 8;
        wValue |= 0x0006;

        for (i = 0; i < COMMAND_RETRY_TIMES; i++)
        {
            ret = SendVendorCommand(m_fdev[iHubIndex],  // device handle
                                    false,              // USB_DIR
                                    GL_ISP_WRITE,       // request
                                    wValue,             // value
                                    0x0000,             // index
                                    0x0000,             // len
                                    NULL);              // buf
            if (ret >= 0)
            {
                break;
            }
            else
            {
                usleep(USLEEP_TIME);
            }
        }

        if (ret < 0)
            goto exit;
    }

    // Re-enter ISP mode
    if (!ModifyHubIspMode(iHubIndex, 0))
    {
        printf("Enter ISP mode failed!\n");
        goto exit;
    }

    status = true;
exit:
    return status;
}

bool SoftwareReset(int iHubIndex)
{
    bool status = true;
    int ret;

    ret = SendVendorCommand(m_fdev[iHubIndex],  // device handle
                            false,              // USB_DIR
                            GL_ISP_SW,          // request
                            0x0003,             // value
                            0x0000,             // index
                            0x0000,             // len
                            NULL);              // buf

    if (ret < 0)
    {
        // printf("SoftwareReset Error!\n");
    }

    return status;
}

bool SwitchUsbSpeed(int iHubIndex, int iUsbSpeed)
{
    bool status         = true;
    unsigned int wValue = 0x0000;
    int ret;

    if (iUsbSpeed == 2)
    {
        wValue = 0x0005;
    }
    else if (iUsbSpeed == 3)
    {
        wValue = 0x0006;
    }

    ret = SendVendorCommand(m_fdev[iHubIndex],  // device handle
                            false,              // USB_DIR
                            GL_ISP_SW,          // request
                            wValue,             // value
                            0x0000,             // index
                            0x0000,             // len
                            NULL);              // buf

    if (ret < 0)
    {
        // printf("SwitchUsbSpeed Error!\n");
    }

    return status;
}

static bool EraseFlash(int iHubIndex, int flashidx, bool reOpen)
{
    bool status       = false;
    unsigned long tmp = 0;
    unsigned int startctTime, tsTime;
    int ret = 0;

    if (reOpen && !OpenDevices())
    {
        printf("Device or driver is not found!");
        status = false;
        goto exit;
    }

    tmp = m_flashinfo[flashidx][FLASH_INFO_ERASE_COMMAND];
    tmp <<= 8;
    tmp += 1;

    ret = SendVendorCommand(m_fdev[iHubIndex],  // device handle
                            false,              // USB_DIR
                            GL_ISP_WRITE,       // request
                            tmp,                // value
                            0x0000,             // index
                            0x0000,             // len
                            NULL);              // buf

    if (ret < 0)
    {
        printf("EraseFlash SendVendorCommand fail\n");
        goto exit;
    }

    startctTime = clock();
    tsTime      = 0;

    while (PollingRDSRStatus(iHubIndex))
    {
        if (tsTime > m_iEraseDelay[iHubIndex] && tsTime > 1500)
            goto exit;

        usleep(50000);
        tsTime = (clock() - startctTime);
    }

    status = true;
exit:
    return status;
}

bool CreateIdBuffer(int iChipID)
{
    bool status = false;

    srand(time(NULL));
    memset(m_containerID, 0, sizeof(m_containerID));

    if (true)
    {
        switch (iChipID)
        {
            case CHIP_ID_UNKNOWN:
                // Assign 16 WORD Container ID buffer from GUID
                m_containerID[1]  = (char)rand();
                m_containerID[3]  = (char)rand();
                m_containerID[5]  = (char)rand();
                m_containerID[7]  = (char)rand();
                m_containerID[9]  = (char)rand();
                m_containerID[11] = (char)rand();
                m_containerID[13] = (char)rand();
                m_containerID[15] = (char)rand();
                m_containerID[17] = (char)rand();
                m_containerID[19] = (char)rand();
                m_containerID[21] = (char)rand();
                m_containerID[23] = (char)rand();
                m_containerID[25] = (char)rand();
                m_containerID[27] = (char)rand();
                m_containerID[29] = (char)rand();
                m_containerID[31] = (char)rand();
                break;

            case CHIP_ID_GL352X_21_SPI:
            case CHIP_ID_GL352X_21_EEPROM:
            case CHIP_ID_GL352X_21_EEPROM_LITE:
                // Assign 16 BYTE Container ID buffer from GUID
                m_containerID[0]  = (char)rand();
                m_containerID[1]  = (char)rand();
                m_containerID[2]  = (char)rand();
                m_containerID[3]  = (char)rand();
                m_containerID[4]  = (char)rand();
                m_containerID[5]  = (char)rand();
                m_containerID[6]  = (char)rand();
                m_containerID[7]  = (char)rand();
                m_containerID[8]  = (char)rand();
                m_containerID[9]  = (char)rand();
                m_containerID[10] = (char)rand();
                m_containerID[11] = (char)rand();
                m_containerID[12] = (char)rand();
                m_containerID[13] = (char)rand();
                m_containerID[14] = (char)rand();
                m_containerID[15] = (char)rand();
                break;

            default:
                break;
        }
        status = true;
        goto exit;
    }
    else
    {
        status = false;
        goto exit;
    }

    status = true;
exit:
    return status;
}

int readBinFileFromIni(int iHubIndex)
{
    char tmpbinFileName[MAX_BUF_LEN] = {0};
    char HubSection[MAX_BUF_LEN]     = {0};

    memset(tmpbinFileName, 0, sizeof(tmpbinFileName));
    sprintf(HubSection, "%s%d", HUB_SECTION, iHubIndex + 1);

    strncpy(tmpbinFileName, (char*)GetIniKeyString(HubSection, BIN_FILE, m_configFilePath), MAX_BUF_LEN);
    if (strcmp(tmpbinFileName, "") == 0)
    {
        printf("Hub%d binFileName is empty!\n", iHubIndex + 1);
        return false;
    }

    return readBinFileName(iHubIndex, tmpbinFileName);
}

int readBinFileName(int iHubIndex, char* binFileName)
{
    int status                       = false;
    char* trimSpaceTmpBinFileName    = NULL;
    char tmpbinFileName[MAX_BUF_LEN] = {0};
    char HubbinFileName[MAX_BUF_LEN] = {0};
    FILE* fp                         = NULL;

    memset(tmpbinFileName, 0, sizeof(tmpbinFileName));
    memset(HubbinFileName, 0, sizeof(HubbinFileName));

    strcpy(tmpbinFileName, binFileName);

    trimSpaceTmpBinFileName = trimwhitespace(tmpbinFileName);
    GetCurrentPath(HubbinFileName, trimSpaceTmpBinFileName);

    // open bin file for reading.
    fp = fopen(HubbinFileName, "rb");
    if (fp == NULL)
    {
        printf("Hub(%d) binFile %s open fail, might be non-exist!\n", iHubIndex + 1, tmpbinFileName);
        status = false;
        goto exit;
    }

    // get file size
    fseek(fp, 0L, SEEK_END);
    m_2BinSize[iHubIndex] = ftell(fp);
    fseek(fp, 0L, SEEK_SET);

    // calloc memory
    m_pBinBuf = (char*)calloc(1, m_2BinSize[iHubIndex]);  // remember to free me
    fread(m_pBinBuf, m_2BinSize[iHubIndex], 1, fp);
    fclose(fp);

    m_BinType[iHubIndex] = GetBinChipID(iHubIndex, m_pBinBuf, m_2BinSize[iHubIndex]);
    status               = true;
exit:
    return status;
}

int Return3523Bonding(GL_STRING_DESCRIPTOR glStr)
{
    int matchBonding = 0;

    if (memcmp(glStr.StaticStr.MaskID, GL_7394_A01, 7) == 0)
    {
        matchBonding = GL_BONDING_7394_A01;
    }
    else if (memcmp(glStr.StaticStr.MaskID, GL_7394_B02, 7) == 0)
    {
        matchBonding = GL_BONDING_7394_B02;
    }
    else if (memcmp(glStr.StaticStr.MaskID, GL_7394_C03, 7) == 0)
    {
        matchBonding = GL_BONDING_7394_C03;
    }
    else if (memcmp(glStr.StaticStr.MaskID, GL_7436_A01, 7) == 0)
    {
        matchBonding = GL_BONDING_7436_A01;
    }
    else if (memcmp(glStr.StaticStr.MaskID, GL_7436_B01, 7) == 0)
    {
        matchBonding = GL_BONDING_7436_B01;
    }
    else if (memcmp(glStr.StaticStr.MaskID, GL_7436_C02, 7) == 0)
    {
        matchBonding = GL_BONDING_7436_C02;
    }
    else if (memcmp(glStr.StaticStr.MaskID, GL_7436_D03, 7) == 0)
    {
        matchBonding = GL_BONDING_7436_D03;
    }
    else if (memcmp(glStr.StaticStr.MaskID, GL_7441_A01, 7) == 0)
    {
        matchBonding = GL_BONDING_7441_A01;
    }
    else if (memcmp(glStr.StaticStr.MaskID, GL_7441_B02, 7) == 0)
    {
        matchBonding = GL_BONDING_7441_B02;
    }
    else if (memcmp(glStr.StaticStr.MaskID, GL_7441_C03, 7) == 0)
    {
        matchBonding = GL_BONDING_7441_C03;
    }
    else if (memcmp(glStr.StaticStr.MaskID, GL_7441_D04, 7) == 0)
    {
        matchBonding = GL_BONDING_7441_D04;
    }

    return matchBonding;
}

int Return3523BinBonding(PGL_SS_HUB_BIN_INFO pGlBinInfo)
{
    int matchBonding = 0;

    if (strstr((char*)pGlBinInfo->ChipInfo.ChiplVersion, GL_7394_A01))
    {
        matchBonding = GL_BONDING_7394_A01;
    }
    else if (strstr((char*)pGlBinInfo->ChipInfo.ChiplVersion, GL_7394_B02))
    {
        matchBonding = GL_BONDING_7394_B02;
    }
    else if (strstr((char*)pGlBinInfo->ChipInfo.ChiplVersion, GL_7394_C03))
    {
        matchBonding = GL_BONDING_7394_C03;
    }
    else if (strstr((char*)pGlBinInfo->ChipInfo.ChiplVersion, GL_7436_A01))
    {
        matchBonding = GL_BONDING_7436_A01;
    }
    else if (strstr((char*)pGlBinInfo->ChipInfo.ChiplVersion, GL_7436_B01))
    {
        matchBonding = GL_BONDING_7436_B01;
    }
    else if (strstr((char*)pGlBinInfo->ChipInfo.ChiplVersion, GL_7436_C02))
    {
        matchBonding = GL_BONDING_7436_C02;
    }
    else if (strstr((char*)pGlBinInfo->ChipInfo.ChiplVersion, GL_7436_D03))
    {
        matchBonding = GL_BONDING_7436_D03;
    }
    else if (strstr((char*)pGlBinInfo->ChipInfo.ChiplVersion, GL_7441_A01))
    {
        matchBonding = GL_BONDING_7441_A01;
    }
    else if (strstr((char*)pGlBinInfo->ChipInfo.ChiplVersion, GL_7441_B02))
    {
        matchBonding = GL_BONDING_7441_B02;
    }
    else if (strstr((char*)pGlBinInfo->ChipInfo.ChiplVersion, GL_7441_C03))
    {
        matchBonding = GL_BONDING_7441_C03;
    }
    else if (strstr((char*)pGlBinInfo->ChipInfo.ChiplVersion, GL_7441_D04))
    {
        matchBonding = GL_BONDING_7441_D04;
    }

    return matchBonding;
}

int ReturnMaskCodeChipID(GL_USB_DEVINFO DevInfo, GL_STRING_DESCRIPTOR glStr, int iHubIndex)
{
    int ret = 0;

    if (DevInfo.bcdDev >= GL7364_MASK_CODE_FW_VERSION_164 && DevInfo.bcdDev < GL7344_MASK_CODE_FW_VERSION)
    {
        ret = CHIP_ID_GL3510_SPI;
    }
    else if (DevInfo.bcdDev >= GL7344_MASK_CODE_FW_VERSION && DevInfo.bcdDev < GL7364_MASK_CODE_FW_VERSION)
    {
        ret = CHIP_ID_GL352X_10_SPI;
    }
    else if (DevInfo.bcdDev >= GL7364_MASK_CODE_FW_VERSION && DevInfo.bcdDev < GL7394_MASK_CODE_FW_VERSION_500)
    {
        ret = CHIP_ID_GL352X_21_SPI;
    }
    else if (DevInfo.bcdDev >= GL7394_MASK_CODE_FW_VERSION_500 && DevInfo.bcdDev < GL8423_MASK_CODE_FW_VERSION_600)
    {
        int hex_val  = -1;
        bool b7394_A = false;

        hex_val = ASCIIConvert(glStr.DynamicStr.Bonding1);
        // hex_val = 0x09;
        if (hex_val == 0)
        {
            // 7394 A01 ver. (Tool string version 0)
            ret     = CHIP_ID_GL3523_SPI;
            b7394_A = true;
        }

        if (b7394_A == false)
        {
            if (glStr.StaticStr.StringVersion <= TOOL_STRING_VERSION_1)
            {
                if (hex_val & BIT_OFFSET_2)
                    ret = CHIP_ID_GL3523_SPI;
                else
                    ret = CHIP_ID_GL3523_TYPE_C;
            }
            else if (glStr.StaticStr.StringVersion >= TOOL_STRING_VERSION_2)
            {
                if ((hex_val & BIT_OFFSET_0) == 0)
                    ret = CHIP_ID_GL3523_SUPPORT_QC;
                else if (hex_val & BIT_OFFSET_3)
                    ret = CHIP_ID_GL3523_SPI;
                else
                    ret = CHIP_ID_GL3523_TYPE_C;
            }
        }
    }
    else if (DevInfo.bcdDev >= GL8423_MASK_CODE_FW_VERSION_600 && DevInfo.bcdDev < GL7448A001_MASK_CODE_FW_VERSION_628)
    {
        ret = CHIP_ID_GL3590_SPI;
    }
    else if (DevInfo.bcdDev >= GL7448A001_MASK_CODE_FW_VERSION_628 && DevInfo.bcdDev < GL7436_MASK_CODE_FW_VERSION_650)
    {
        ret = CHIP_ID_GL7000_SPI;
    }
    else if (DevInfo.bcdDev >= GL7436_MASK_CODE_FW_VERSION_650)
    {
        ret = CHIP_ID_GL3510_SPI;
    }

    return ret;
}
