/******************************************************************************
 * cliArg.c
 *
 *
 *****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <getopt.h>

#include "cliArg.h"

/*
 * prepare opt_string for getopt_long()
 */
static void build_opt_string(char *opt_string, struct arg_handle *arg_handles, int arg_num)
{
    int b = 0;
    int i;

    for (i = 0; i < arg_num; i++)
    {
        opt_string[b++] = arg_handles[i].short_name;

        if (arg_handles[i].type != ARG_TYPE_NONE)
        {
            opt_string[b++] = ':';
        }
    }

    opt_string[b] = '\0';

    return;
}

/*
 * prepare longopts for getopt_long()
 */
static void build_long_opts(struct option *long_opts, struct arg_handle *arg_handles, int arg_num)
{
    int i;

    for (i = 0; i < arg_num; i++)
    {
        long_opts[i].name    = arg_handles[i].long_name;
        long_opts[i].has_arg = (arg_handles[i].type != ARG_TYPE_NONE);
        long_opts[i].flag    = NULL;
        long_opts[i].val     = (int)arg_handles[i].short_name;
    }

    /* end longopts array with all zero */
    long_opts[arg_num].name    = NULL;
    long_opts[arg_num].has_arg = 0;
    long_opts[arg_num].flag    = NULL;
    long_opts[arg_num].val     = 0;

    return;
}

/*
 * get integer argument
 */
static int arg_get_integer(char *arg, int *integer)
{
    if (strncmp(arg, "0x", 2) == 0)
    {
        errno    = 0;
        *integer = (int)strtol(arg, NULL, 16);
    }
    else
    {
        errno    = 0;
        *integer = (int)strtol(arg, NULL, 10);
    }

    if (errno != 0)
    {
        return EINVAL;
    }

    return errno;
}

/*
 * get string argument
 */
static int arg_get_string(char *arg, char **pstr)
{
    char *p = arg;
    char *string;

    *pstr = NULL;

    if ((string = (char *)malloc(strlen(p) + 1)) == NULL)
    {
        return ENOMEM;
    }

    if (strlen(p) > 0)
    {
        strncpy(string, p, strlen(p));
    }

    string[strlen(p)] = 0;

    *pstr = string;

    return 0;
}

/*
 * print argument usage, print to stdout if fp == NULL
 */
void arg_print_usage(struct arg_handle *arg_handles, int arg_num, FILE *fp)
{
    char buf1[1024];
    char tmpb[256];
    int opt, i;

    for (opt = 0; opt < 2; opt++)
    {
        for (i = 0; i < arg_num; i++)
        {
            if (arg_handles[i].optional != opt)
            {
                continue;
            }

            buf1[0] = '\0';

            if (arg_handles[i].optional != 0)
            {
                strcat(buf1, "[");
            }
            snprintf(tmpb, 256, "-%c%s", arg_handles[i].short_name,
                     (arg_handles[i].type != ARG_TYPE_NONE ? " <arg>" : ""));
            strcat(buf1, tmpb);

            strcat(buf1, "|");
            snprintf(tmpb, 256, "--%s%s", arg_handles[i].long_name,
                     (arg_handles[i].type != ARG_TYPE_NONE ? " <arg>" : ""));

            strcat(buf1, tmpb);

            if (arg_handles[i].optional != 0)
            {
                strcat(buf1, "]");
            }

            strcat(buf1, " : ");
            strcat(buf1, arg_handles[i].description);

            fprintf((fp == NULL ? stdout : fp), "\t%s\n", buf1);
        }
    }
}

/*
 * parse arguments
 */
int arg_get_arguments(struct arg_handle *arg_handles, int arg_num, int argc, char *argv[])
{
    char opt_string[64];
    struct option long_opts[32];
    int longidx;
    int ret;
    char c;
    int i;

    /* get parameters */
    optind = 1;
    build_opt_string(opt_string, arg_handles, arg_num);
    build_long_opts(long_opts, arg_handles, arg_num);

    while ((ret = getopt_long(argc, argv, opt_string, long_opts, &longidx)) > 0)
    {
        c = (char)ret;

        for (i = 0; i < arg_num; i++)
        {
            if (c == arg_handles[i].short_name)
            {
                if (arg_handles[i].type != ARG_TYPE_NONE)
                {
                    /* argument is missing */
                    if (strcmp(optarg, ":") == 0 || strcmp(optarg, "?") == 0)
                    {
                        return EFAULT;
                    }

                    /* convert argument to integer or string */
                    switch (arg_handles[i].type)
                    {
                        case ARG_TYPE_INT:
                            if (arg_get_integer(optarg, &(arg_handles[i].val.integer)) != 0)
                            {
                                return EINVAL;
                            }
                            break;

                        case ARG_TYPE_STRING:
                            if (arg_get_string(optarg, &(arg_handles[i].val.string)) != 0)
                            {
                                return EINVAL;
                            }
                            break;

                        default:
                            return EINVAL;
                    }
                }

                /* parameter already set */
                if (arg_handles[i].is_set == 1)
                {
                    return EEXIST;
                }

                arg_handles[i].is_set = 1;
                break;
            }
        }

        if (i >= arg_num)
        {
            return ENOENT;
        }
    }

    /* make sure madetory arguments are all set */
    /* ignore checking for command 'dismiss' and 'protection_status' that is checked in its specific function */
    if (strcmp(argv[0], "dismiss") != 0 && strcmp(argv[0], "protection_status") != 0)
    {
        for (i = 0; i < arg_num; i++)
        {
            if (arg_handles[i].optional == 0 && arg_handles[i].is_set == 0)
            {
                return ENODATA;
            }
        }
    }

    return 0;
}

/*
 * release memory allocated for arguments
 */
void arg_release_arguments(struct arg_handle *arg_handles, int arg_num)
{
    int i;

    for (i = 0; i < arg_num; i++)
    {
        if (arg_handles[i].type == ARG_TYPE_STRING)
        {
            if (arg_handles[i].val.string != NULL)
            {
                free(arg_handles[i].val.string);
                arg_handles[i].val.string = NULL;
            }
        }
        arg_handles[i].is_set = 0;
    }
}

/* vim: set tabstop=4 shiftwidth=4 softtabstop=4 expandtab: */
