/******************************************************************************
 * hubFwUpdaterCLI.c
 *
 * hubFwUpdaterCLI main function
 *****************************************************************************/
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include "hubFwUpdaterCLI.h"
#include "cliFunc.h"

struct cli_command
{
    char name[MAX_CMD_LEN];
    char description[MAX_BUF_LEN];
    int (*cmd_func)(int argc, char *argv[]);
};

struct cli_command cli_commands[] = {
    {"help", "help", help_func},
    {"isp", "isp firmware", isp_firmware},
    {"erase", "erase firmware", erase_firmware},
    {"version", "firmware version", read_fw_version},
    {"tree", "list hub tree", list_hub_tree},
    {"reset", "software reset", software_reset},
    {"switch_speed", "switch usb speed", switch_speed},
    {"info", "list hub info", list_info},
};

int cli_command_num = sizeof(cli_commands) / sizeof(struct cli_command);
char cli_result[MAX_BUF_LEN];

void cli_usage(char *prog)
{
    int cmd = 0;

    printf("Genesys hubFwUpdaterCLI version %s\n", VERSION);
    printf("List of commands:\n");

    for (cmd = 0; cmd < cli_command_num; cmd++)
    {
        if (cli_commands[cmd].cmd_func != NULL)
        {
            printf("\t%s -- %s\n", cli_commands[cmd].name, cli_commands[cmd].description);
        }
    }
    printf("\n");
}

int remove_verbose(int in_argc, char *in_argv[], int out_argc, char *out_argv[])
{
    int is_verbose = 0;
    int i, j;

    for (i = 0, j = 0; i < in_argc && j < out_argc; i++)
    {
        if (strcmp(in_argv[i], "-v") != 0 && strcmp(in_argv[i], "--verbose") != 0)
        {
            out_argv[j] = (char *)calloc(strlen(in_argv[i]) + 1, sizeof(char));
            strncpy(out_argv[j], in_argv[i], strlen(in_argv[i]));
            j++;
        }
        else
        {
            is_verbose++;
        }
    }

    return is_verbose;
}

int main(int argc, char *argv[])
{
    int cmd         = 0;
    int retcode     = 0;
    int verbose     = 0;
    int cmd_argc    = 0;
    char **cmd_argv = NULL;

    if (argc < 2)
    {
        cli_usage(argv[0]);
        return EINVAL;
    }

    cmd_argv = (char **)calloc(argc, sizeof(char *));
    cmd_argc = argc;
    /* '-v or --verbose' is global argument of drmcli for printing command summary
     * do not pass this argumet into each command function */
    verbose = remove_verbose(argc, argv, cmd_argc, cmd_argv);
    /* verbose mode is set, subtract the verbose count from the argument count
     * usually, the verbose count should be 1 but in case someone set both -v and --verbose */
    if (verbose > 0)
    {
        cmd_argc = cmd_argc - verbose;
    }

    for (cmd = 0; cmd < cli_command_num; cmd++)
    {
        if (cli_commands[cmd].cmd_func != NULL)
        {
            if (strcmp(argv[1], cli_commands[cmd].name) == 0)
            {
                /* run command */

                memset(cli_result, 0, sizeof(cli_result));
                if ((retcode = cli_commands[cmd].cmd_func(cmd_argc, cmd_argv)) != 0)
                {
                    if (retcode != ENODATA)
                    {
                        if (retcode == ENAVAIL)
                        {
                            cli_usage(argv[0]);
                            retcode = 0;
                        }
                        else
                        {
                            if (verbose > 0)
                            {
                                /* print standard message if command return empty result */
                                if (strlen(cli_result) == 0)
                                {
                                    fprintf(stderr, "\nCommand [%s] returns error:%s.\n", cli_commands[cmd].name,
                                            errno_shortname[retcode]);
                                }
                                else
                                {
                                    fprintf(stderr, "\n%s\n", cli_result);
                                }
                            }
                            else
                            {
                                fprintf(stderr, "\nCommand [%s] returns error:%s\n", cli_commands[cmd].name,
                                        errno_shortname[retcode]);
                            }
                        }
                    }
                }
                else
                {
                    if (verbose > 0)
                    {
                        /* print standard message if command return empty result */
                        if (strlen(cli_result) == 0)
                        {
                            fprintf(stderr, "\nCommand '%s' executed successfully.\n", cli_commands[cmd].name);
                        }
                        else
                        {
                            fprintf(stderr, "\n%s\n", cli_result);
                        }
                    }
                }

                break;
            }
        }
    }

    /* unknown command? */
    if (cmd == cli_command_num)
    {
        cli_usage(argv[0]);
        return EINVAL;
    }

    /* free cmd_argv */
    if (cmd_argv != NULL)
    {
        for (cmd_argc = 0; cmd_argc < argc; cmd_argc++)
        {
            if (cmd_argv[cmd_argc] != NULL)
            {
                free(cmd_argv[cmd_argc]);
            }
        }
        free(cmd_argv);
    }

    return retcode;
}
